/*
 * Copyright 2020 Free Software Foundation, Inc.
 *
 * This file is part of GNU Radio
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 */
#include "pydoc_macros.h"
#define D(...) DOC(osmosdr, __VA_ARGS__ )
/*
  This file contains placeholders for docstrings for the Python bindings.
  Do not edit! These were automatically extracted during the binding process
  and will be overwritten during the build process
 */


 
 static const char *__doc_osmosdr_sink = R"doc()doc";


 static const char *__doc_osmosdr_sink_sink_0 = R"doc()doc";


 static const char *__doc_osmosdr_sink_sink_1 = R"doc()doc";


 static const char *__doc_osmosdr_sink_make = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_num_channels = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_sample_rates = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_sample_rate = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_sample_rate = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_freq_range = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_center_freq = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_center_freq = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_freq_corr = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_freq_corr = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_gain_names = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_gain_range_0 = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_gain_range_1 = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_gain_mode = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_gain_mode = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_gain_0 = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_gain_1 = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_gain_0 = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_gain_1 = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_if_gain = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_bb_gain = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_antennas = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_antenna = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_antenna = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_dc_offset = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_iq_balance = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_bandwidth = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_bandwidth = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_bandwidth_range = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_time_source = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_time_source = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_time_sources = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_clock_source = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_clock_source = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_clock_sources = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_clock_rate = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_clock_rate = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_time_now = R"doc()doc";


 static const char *__doc_osmosdr_sink_get_time_last_pps = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_time_now = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_time_next_pps = R"doc()doc";


 static const char *__doc_osmosdr_sink_set_time_unknown_pps = R"doc()doc";

  
